package cz.fidentis.analyst.symmetry;

import cz.fidentis.analyst.scene.DrawableFace;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.util.List;
import javax.swing.JPanel;
import javax.vecmath.Point3d;


/**
 * Panel for graphing polyline profiles
 *
 * @author Dominik Racek
 * @author Radek Oslejsek
 */
public class CurveRenderingPanel extends JPanel {
    
    private static final int PREF_W = 400;
    private static final int PREF_H = 400;
    private static final int BORDER_GAP = 15;
    private static final Color PRIMARY_COLOR = DrawableFace.SKIN_COLOR_PRIMARY;
    private static final Color PRIMARY_MIRROR_COLOR = PRIMARY_COLOR.darker();
    private static final Color SECONDARY_COLOR = DrawableFace.SKIN_COLOR_SECONDARY.darker();
    private static final Color SECONDARY_MIRROR_COLOR = SECONDARY_COLOR.darker().darker();
    /*
    private static final Color PRIMARY_COLOR = Color.green;
    private static final Color PRIMARY_MIRROR_COLOR = new Color(0, 85, 0);
    private static final Color SECONDARY_COLOR = new Color(0, 155, 200);
    private static final Color SECONDARY_MIRROR_COLOR = Color.blue;
    */
    private static final Stroke GRAPH_STROKE = new BasicStroke(3f);
    
    private List<List<Point3d>> primarySegments;
    private List<List<Point3d>> secondarySegments;
    private List<List<Point3d>> primaryMirrorSegments;
    private List<List<Point3d>> secondaryMirrorSegments;
    
    private double primaryOffsetZ = 0;
    private boolean mirrorCuts = false;
    
    private double minZ = Double.POSITIVE_INFINITY;
    private double maxZ = Double.NEGATIVE_INFINITY;
    private double minY = Double.POSITIVE_INFINITY;
    private double maxY = Double.NEGATIVE_INFINITY;
    private double scale = Double.POSITIVE_INFINITY;
    
    /**
     * Constructor.
     */
    public CurveRenderingPanel() {
        setPreferredSize(new Dimension(PREF_W, PREF_H));
    }

    /**
     * Sets primary points and draws them in the panel
     *
     * @param segments primary points
     */
    public void setPrimarySegments(List<List<Point3d>> segments) {
        this.primarySegments = segments;
        updateMinMax(segments);
        repaint();
    }

    /**
     * Sets primary mirror points.
     * Only draws them if mirrorCuts is checked
     *
     * @param segments primary mirror points
     */
    public void setPrimaryMirrorSegments(List<List<Point3d>> segments) {
        this.primaryMirrorSegments = segments;
        updateMinMax(segments);
        repaint();
    }

    /**
     * Sets secondary points and draws them in the panel
     *
     * @param segments secondary points
     */
    public void setSecondarySegments(List<List<Point3d>> segments) {
        this.secondarySegments = segments;
        updateMinMax(segments);
        repaint();
    }

    /**
     * Sets secondary mirror points.
     * Only draws them if mirrorCuts is checked
     *
     * @param segments secondary mirror points
     */
    public void setSecondaryMirrorSegments(List<List<Point3d>> segments) {
        this.secondaryMirrorSegments = segments;
        updateMinMax(segments);
        repaint();
    }

    /**
     * Displays the mirror cuts
     *
     * @param mirror
     */
    public void setMirrorCuts(boolean mirror) {
        this.mirrorCuts = mirror;
        repaint();
    }

    @Override
    public Dimension getPreferredSize() {
        return new Dimension(PREF_W, PREF_H);
    }

    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        Graphics2D g2 = (Graphics2D)g;

        if (mirrorCuts) {
            drawFace(g2, primarySegments, PRIMARY_COLOR, true);
            drawFace(g2, primaryMirrorSegments, PRIMARY_MIRROR_COLOR, true);

            if (secondarySegments != null) {
                drawFace(g2, secondarySegments, SECONDARY_COLOR, false);
                drawFace(g2, secondaryMirrorSegments, SECONDARY_MIRROR_COLOR, false);
            }
        } else {
            drawFace(g2, primarySegments, PRIMARY_COLOR, true);

            if (secondarySegments != null) {
                drawFace(g2, secondarySegments, SECONDARY_COLOR, false);
            }
        }
    }

    protected void drawFace(Graphics2D g2, List<List<Point3d>> curve, Color faceColor, boolean isPrimary) {
        if (curve == null) {
            return;
        }
        
        for (int i = 0; i < curve.size(); i++) {
            drawCurveSegment(g2, curve.get(i), faceColor, isPrimary, maxY, maxZ);
        }
    }

    protected void drawCurveSegment(Graphics2D g2, List<Point3d> curveSegment, Color faceColor, boolean isPrimary, double maxY, double maxZ) {
        double offsetZ = 0;

        //Calculate the offsets
        if (mirrorCuts) {
            if (secondarySegments == null) {
                //Mirror cuts with single face - center the face
                this.primaryOffsetZ = (PREF_W * 0.7) - (maxZ * scale);
                offsetZ = this.primaryOffsetZ;
            } else {
                //Mirror cuts with two faces - separate primary and secondary
                offsetZ = isPrimary ? (PREF_W * 0.4) - (maxZ * scale) : (PREF_W * 0.9) - (maxZ * scale);
            }
        } else {
            //No mirror cuts - center all faces
            if (isPrimary) {
                this.primaryOffsetZ = (PREF_W * 0.7) - (maxZ * scale);
            }
           //offsetZ = this.alignProfiles ? (PREF_W * 0.7) - (maxZ * scale) : this.primaryOffsetZ;
           offsetZ = this.primaryOffsetZ;
        }

        //Draw lines
        g2.setColor(faceColor);
        g2.setStroke(GRAPH_STROKE);
        for (int i = 0; i < curveSegment.size() - 1; i++) {
            double z1 = (curveSegment.get(i).z) * scale + BORDER_GAP + offsetZ;
            //double y1 = (maxY - curveSegment.get(i).y) * scale + BORDER_GAP;
            double y1 = PREF_H / 2 - curveSegment.get(i).y * scale + 3 * BORDER_GAP;
            double z2 = (curveSegment.get(i + 1).z) * scale + BORDER_GAP + offsetZ;
            //double y2 = (maxY - curveSegment.get(i + 1).y) * scale + BORDER_GAP;
            double y2 = PREF_H / 2 - curveSegment.get(i + 1).y * scale + 3 * BORDER_GAP;
            
            g2.draw(new Line2D.Double(z1, y1, z2, y2));
        }
    }

    protected void updateMinMax(List<List<Point3d>> curve) {
        if (curve == null) {
            return;
        }
        
        for (int i = 0; i < curve.size(); i++) {
            for (int j = 0; j < curve.get(i).size(); j++) {
                minZ = Math.min(minZ, curve.get(i).get(j).z);
                maxZ = Math.max(maxZ, curve.get(i).get(j).z);
                minY = Math.min(minY, curve.get(i).get(j).y);
                maxY = Math.max(maxY, curve.get(i).get(j).y);
            }
        }

        scale = ((double) PREF_H - 2 * BORDER_GAP) / (maxY - minY);
    }
}