package cz.fidentis.analyst.symmetry;

import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.core.ComboSliderDouble;
import cz.fidentis.analyst.core.ControlPanelAction;
import cz.fidentis.analyst.face.events.HumanFaceEvent;
import cz.fidentis.analyst.face.events.HumanFaceListener;
import cz.fidentis.analyst.mesh.core.CornerTableRow;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshFacetImpl;
import cz.fidentis.analyst.mesh.core.MeshPointImpl;
import cz.fidentis.analyst.scene.DrawablePlane;
import cz.fidentis.analyst.visitors.mesh.BoundingBox;
import cz.fidentis.analyst.visitors.mesh.CrossSectionZigZag;
import org.openide.filesystems.FileChooserBuilder;

import javax.swing.JTabbedPane;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.vecmath.Point3d;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.List;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.vecmath.Vector3d;

/**
 * Action listener for the manipulation with the cutting plane.
 *
 * @author Dominik Racek
 * @author Radek Oslejsek
 */
public class ProfilesAction extends ControlPanelAction implements HumanFaceListener {

    /*
     * GUI elements
     */
    private final ProfilesPanel controlPanel;
    private final JTabbedPane topControlPanel;

    /*
     * Calculated profiles
     */
    private List<List<Point3d>> primarySegments;
    private List<List<Point3d>> secondarySegments;
    private List<List<Point3d>> primaryMirrorSegments;
    private List<List<Point3d>> secondaryMirrorSegments;

    private double lastSliderValue = 0.5;

    /**
     * Constructor.
     *
     * @param canvas          OpenGL canvas
     * @param topControlPanel Top component for placing control panels
     */
    public ProfilesAction(Canvas canvas, JTabbedPane topControlPanel) {
        super(canvas, topControlPanel);
        this.topControlPanel = topControlPanel;
        
        computeOrthogonalCuttingPlanes(false);

        //recomputePrimaryProfile();

        if (getSecondaryDrawableFace() != null) {
            //recomputeSecondaryProfile();
            controlPanel = new ProfilesPanel(this, this.primarySegments, this.secondarySegments);
            controlPanel.getProfileRenderingPanel().setSecondaryMirrorSegments(this.secondaryMirrorSegments);
        } else {
            controlPanel = new ProfilesPanel(this, this.primarySegments);
        }
        controlPanel.getProfileRenderingPanel().setPrimaryMirrorSegments(this.primaryMirrorSegments);

        // Place control panel to the topControlPanel
        this.topControlPanel.addTab(controlPanel.getName(), controlPanel.getIcon(), controlPanel);
        this.topControlPanel.addChangeListener(e -> {
            // If the symmetry panel is focused...
            if (((JTabbedPane) e.getSource()).getSelectedComponent() instanceof ProfilesPanel) {
                getCanvas().getScene().setDefaultColors();
                getCanvas().getScene().showCuttingPlanes();
                if (controlPanel.isMirrorCutsChecked()) {
                    getCanvas().getScene().showMirrorPlanes();
                }
                recomputeProfiles();
            } else {
                getCanvas().getScene().hideCuttingPlanes();
                getCanvas().getScene().hideMirrorPlanes();
            }
        });
        
        getCanvas().getScene().hideCuttingPlanes();
        getCanvas().getScene().hideMirrorPlanes();
        
        // Be informed about changes in faces perfomed by other GUI elements
        getPrimaryDrawableFace().getHumanFace().registerListener(this);
        if (getSecondaryDrawableFace() != null) {
            getSecondaryDrawableFace().getHumanFace().registerListener(this);
        }
    }

    @Override
    public void actionPerformed(ActionEvent ae) {
        String action = ae.getActionCommand();
        
        switch (action) {
            case ProfilesPanel.ACTION_COMMAND_EXPORT:
                exportProfile(this.primarySegments, "Export primary face profile to file");
                if (controlPanel.isMirrorCutsChecked()) {
                    exportProfile(this.primaryMirrorSegments, "Export primary face mirror profile to file");
                }

                if (getSecondaryDrawableFace() != null) {
                    exportProfile(this.secondarySegments, "Export secondary face profile to file");
                    if (controlPanel.isMirrorCutsChecked()) {
                        exportProfile(this.secondaryMirrorSegments, "Export secondary face mirror profile to file");
                    }
                }
                break;
            case ProfilesPanel.ACTION_OFFSET_CUTTING_PLANE:
                double auxVal = ((ComboSliderDouble) ae.getSource()).getValue();
                double value =  auxVal - lastSliderValue;
                double multiplier = -150;
                
                setCuttingPlaneOffset(0, multiplier * value);

                if (getSecondaryDrawableFace() != null) {
                    setCuttingPlaneOffset(1, multiplier * value);
                }
                recomputeProfiles();
                lastSliderValue = auxVal;
                break;
            case ProfilesPanel.ACTION_CHANGE_CUTTING_PLANE:
                String option = (String)((JComboBox) ae.getSource()).getSelectedItem();
                
                if (option.equals(ProfilesPanel.OPTION_SYMMETRY_PLANE)) {
                    if (getScene().getDrawableSymmetryPlane(0) == null) {
                       JOptionPane.showMessageDialog(new JFrame(), "Compute a symmetry plane at the Symmetry tab first.", "Dialog", JOptionPane.INFORMATION_MESSAGE);
                       controlPanel.setDefaultPlaneSelection();
                       return;
                    }
                    controlPanel.resetSliderSilently();
                    computeCuttingPlanesFromSymmetry();
                } else if (option.equals(ProfilesPanel.OPTION_VERTICAL_PLANE)) {
                    controlPanel.resetSliderSilently();
                    computeOrthogonalCuttingPlanes(false);
                }
                
                if (controlPanel.isMirrorCutsChecked()) {
                    getCanvas().getScene().showMirrorPlanes();
                } else {
                    getCanvas().getScene().hideMirrorPlanes();
                }

                recomputeProfiles();
                break;
            case ProfilesPanel.ACTION_MIRROR_CUTS:
                controlPanel.getProfileRenderingPanel().setMirrorCuts(controlPanel.isMirrorCutsChecked());
                if (controlPanel.isMirrorCutsChecked()) {
                    getCanvas().getScene().showMirrorPlanes();
                } else {
                    getCanvas().getScene().hideMirrorPlanes();
                }
                break;
            default:
                // do nothing
        }
        renderScene();
    }
    
    @Override
    public void acceptEvent(HumanFaceEvent event) {
    }

    private void exportProfile(List<List<Point3d>> segments, String title) {
        File file = new FileChooserBuilder(ProfilesAction.class)
                .setTitle(title)
                .setDefaultWorkingDirectory(new File(System.getProperty("user.home")))
                .setFilesOnly(true)
                .setFileFilter(new FileNameExtensionFilter("csv files (*.csv)", "csv"))
                .setAcceptAllFileFilterUsed(true)
                .showSaveDialog();

        if (file == null) {
            return;
        }

        // If chosen file exists, use the exact file path
        // If chosen file does not exist and does not have an extension, add it
        if (!file.exists()) {
            if (!file.getAbsolutePath().endsWith(".csv")) {
                file = new File(file.getAbsolutePath() + ".csv");
            }
        }

        try {
            file.createNewFile();

            PrintWriter writer = new PrintWriter(file.getAbsoluteFile(), "UTF-8");
            writer.println("N,X-Coordinate,Z-Coordinate");
            for (int i = 0; i < segments.size(); i++) {
                for (int j = 0; j < segments.get(i).size(); ++j) {
                    writer.println(((i+1)*(j+1)) + "," + segments.get(i).get(j).x + "," + segments.get(i).get(j).z);
                }
            }

            writer.close();
        } catch (IOException ex) {
            System.out.println("ERROR writing to a file: " + ex);
        }
    }

    private void recomputePrimaryProfile() {
        //Main profile
        CrossSectionZigZag cs = new CrossSectionZigZag(getScene().getDrawableCuttingPlane(0).getPlane());
        getPrimaryDrawableFace().getModel().compute(cs);
        this.primarySegments = cs.getSegments();

        //Mirror profile
        CrossSectionZigZag mcs = new CrossSectionZigZag(getScene().getDrawableMirrorPlane(0).getPlane());
        getPrimaryDrawableFace().getModel().compute(mcs);
        this.primaryMirrorSegments = mcs.getSegments();
    }

    private void recomputeSecondaryProfile() {
        //Main profile
        CrossSectionZigZag cs = new CrossSectionZigZag(getScene().getDrawableCuttingPlane(1).getPlane());
        getSecondaryDrawableFace().getModel().compute(cs);
        this.secondarySegments = cs.getSegments();

        //Mirror profile
        CrossSectionZigZag mcs = new CrossSectionZigZag(getScene().getDrawableMirrorPlane(1).getPlane());
        getSecondaryDrawableFace().getModel().compute(mcs);
        this.secondaryMirrorSegments = mcs.getSegments();
    }

    private void recomputeProfiles() {
        recomputePrimaryProfile();
        controlPanel.getProfileRenderingPanel().setPrimarySegments(this.primarySegments);
        controlPanel.getProfileRenderingPanel().setPrimaryMirrorSegments(this.primaryMirrorSegments);

        if (getSecondaryDrawableFace() != null) {
            recomputeSecondaryProfile();
            controlPanel.getProfileRenderingPanel().setSecondarySegments(this.secondarySegments);
            controlPanel.getProfileRenderingPanel().setSecondaryMirrorSegments(this.secondaryMirrorSegments);
        }
    }

    private void setCuttingPlaneOffset(int index, double value) {
        //Move cutting planes left and mirror planes right
        //If normal is negative, need to negate value
        if (getScene().getDrawableCuttingPlane(index).getPlane().getNormal().x < 0) {
            getScene().getDrawableCuttingPlane(index).translatePlane(-value);
        } else {
            getScene().getDrawableCuttingPlane(index).translatePlane(value);
        }
        
        if (getScene().getDrawableMirrorPlane(index).getPlane().getNormal().x < 0) {
            getScene().getDrawableMirrorPlane(index).translatePlane(value);
        } else {
            getScene().getDrawableMirrorPlane(index).translatePlane(-value);
        }
    }
    
    /**
     * Creates vertical or horizontal cutting and mirror planes from bounding box.
     * 
     * @param horizontal 
     */
    protected void computeOrthogonalCuttingPlanes(boolean horizontal) {
        DrawablePlane drPlane = getDrawableOrthogonalPlane(horizontal);
        getCanvas().getScene().setDrawableCuttingPlane(0, drPlane);
        getCanvas().getScene().setDrawableMirrorPlane(0, new DrawablePlane(drPlane));
        if (getSecondaryDrawableFace() != null) {
            getCanvas().getScene().setDrawableCuttingPlane(1, drPlane);
            getCanvas().getScene().setDrawableMirrorPlane(1, new DrawablePlane(drPlane));
        }
    }
    
    protected DrawablePlane getDrawableOrthogonalPlane(boolean horizontal) {
        BoundingBox bbox = new BoundingBox();
        getPrimaryDrawableFace().getModel().compute(bbox);
        if (getSecondaryDrawableFace() != null) {
            getSecondaryDrawableFace().getModel().compute(bbox);
        }
        
        Vector3d normal = (horizontal) ? new Vector3d(0,1,0) : new Vector3d(1,0,0);
        Point3d midPoint = bbox.getBoundingBox().getMidPoint();
        Point3d minPoint = bbox.getBoundingBox().getMinPoint();
        Point3d maxPoint = bbox.getBoundingBox().getMaxPoint();
        
        // add border:
        Vector3d diagonal = new Vector3d(maxPoint);
        diagonal.sub(minPoint);
        diagonal.scale(0.15); // border size
        maxPoint.add(diagonal);
        diagonal.scale(-1);
        minPoint.add(diagonal);
        
        Plane plane = new Plane(normal, (horizontal) ? midPoint.y : midPoint.x);
        
        MeshFacet planeMesh = new MeshFacetImpl();
        if (horizontal) {
            planeMesh.addVertex(new MeshPointImpl(new Point3d(minPoint.x, midPoint.y, maxPoint.z), normal, null));
            planeMesh.addVertex(new MeshPointImpl(new Point3d(minPoint.x, midPoint.y, minPoint.z), normal, null));
            planeMesh.addVertex(new MeshPointImpl(new Point3d(maxPoint.x, midPoint.y, minPoint.z), normal, null));
            planeMesh.addVertex(new MeshPointImpl(new Point3d(maxPoint.x, midPoint.y, maxPoint.z), normal, null));
        } else {
            planeMesh.addVertex(new MeshPointImpl(new Point3d(midPoint.x, minPoint.y, minPoint.z), normal, null));
            planeMesh.addVertex(new MeshPointImpl(new Point3d(midPoint.x, maxPoint.y, minPoint.z), normal, null));
            planeMesh.addVertex(new MeshPointImpl(new Point3d(midPoint.x, maxPoint.y, maxPoint.z), normal, null));
            planeMesh.addVertex(new MeshPointImpl(new Point3d(midPoint.x, minPoint.y, maxPoint.z), normal, null));
        }
        planeMesh.getCornerTable().addRow(new CornerTableRow(0, 5));
        planeMesh.getCornerTable().addRow(new CornerTableRow(1, -1));
        planeMesh.getCornerTable().addRow(new CornerTableRow(3, -1));
        planeMesh.getCornerTable().addRow(new CornerTableRow(3, -1));
        planeMesh.getCornerTable().addRow(new CornerTableRow(1, -1));
        planeMesh.getCornerTable().addRow(new CornerTableRow(2, 0));

        DrawablePlane cuttingPlane = new DrawablePlane(planeMesh, plane);
        cuttingPlane.setTransparency(0.5f);
        return cuttingPlane;
    }
    
    /**
     * Copy rectangle from the symmetry plane.
     * 
     * @param faceIndex
     * @return 
     */
    protected DrawablePlane getCuttingPlaneFromSymmetry(int faceIndex) {
        MeshFacet symmetryFacet = getScene().getHumanFace(faceIndex).getSymmetryPlaneFacet();
        Plane symmetryPlane = getScene().getHumanFace(faceIndex).getSymmetryPlane();
        return new DrawablePlane(new MeshFacetImpl(symmetryFacet), new Plane(symmetryPlane));
    }
    
    /**
     * Creates cutting and mirror planes by copying the rectangle from the symmetry plane.
     */
    protected void computeCuttingPlanesFromSymmetry() {
        DrawablePlane cuttingPlane = getCuttingPlaneFromSymmetry(0);
        cuttingPlane.setTransparency(0.5f);
        getCanvas().getScene().setDrawableCuttingPlane(0, cuttingPlane);
        getCanvas().getScene().setDrawableMirrorPlane(0, new DrawablePlane(cuttingPlane));
        recomputePrimaryProfile();
        
        if (getSecondaryDrawableFace() != null) {
            cuttingPlane = getCuttingPlaneFromSymmetry(1);
            cuttingPlane.setTransparency(0.5f);
            getCanvas().getScene().setDrawableCuttingPlane(1, cuttingPlane);
            getCanvas().getScene().setDrawableMirrorPlane(1, new DrawablePlane(cuttingPlane));
            recomputeSecondaryProfile();
        }
    }

}
